@@ -11,7 +11,7 @@
 use super::igd::{forward_port, IgdError};
 use super::wire_msg::WireMsg;
 use super::{
-    config::{Config, InternalConfig, RetryConfig},
+    config::{Config, InternalConfig},
     connection_deduplicator::{ConnectionDeduplicator, DedupHandle},
     connection_pool::{ConnId, ConnectionPool, ConnectionRemover},
     connections::{
@@ -19,7 +19,7 @@ use super::{
         DisconnectionEvents, RecvStream, SendStream,
     },
     error::{
-        ClientEndpointError, ConnectionError, EndpointError, RecvError, RpcError, SendError,
+        ClientEndpointError, ConnectionError, EndpointError, RecvError, RpcError,
         SerializationError,
     },
 };
@@ -46,22 +46,24 @@ const STANDARD_CHANNEL_SIZE: usize = 10000;
 
 /// Channel on which incoming messages can be listened to
 #[derive(Debug)]
-pub struct IncomingMessages(pub(crate) MpscReceiver<(SocketAddr, Bytes)>);
+pub struct IncomingMessages<I: ConnId>(pub(crate) MpscReceiver<(Connection<I>, Bytes)>);
 
-impl IncomingMessages {
-    /// Blocks and returns the next incoming message and the source peer address
-    pub async fn next(&mut self) -> Option<(SocketAddr, Bytes)> {
+impl<I: ConnId> IncomingMessages<I> {
+    /// Blocks and returns the next incoming message and the connection it arrived on.
+    ///
+    /// **Note:** holding on to `Connection`
+    pub async fn next(&mut self) -> Option<(Connection<I>, Bytes)> {
         self.0.recv().await
     }
 }
 
 /// Channel on which incoming connections are notified on
-pub struct IncomingConnections(pub(crate) MpscReceiver<SocketAddr>);
+pub struct IncomingConnections<I: ConnId>(pub(crate) MpscReceiver<Connection<I>>);
 
-impl IncomingConnections {
+impl<I: ConnId> IncomingConnections<I> {
     /// Blocks until there is an incoming connection and returns the address of the
     /// connecting peer
-    pub async fn next(&mut self) -> Option<SocketAddr> {
+    pub async fn next(&mut self) -> Option<Connection<I>> {
         self.0.recv().await
     }
 }
@@ -72,7 +74,7 @@ pub struct Endpoint<I: ConnId> {
     local_addr: SocketAddr,
     public_addr: Option<SocketAddr>,
     quic_endpoint: quinn::Endpoint,
-    message_tx: MpscSender<(SocketAddr, Bytes)>,
+    message_tx: MpscSender<(Connection<I>, Bytes)>,
     disconnection_tx: MpscSender<SocketAddr>,
     config: InternalConfig,
     termination_tx: Sender<()>,
@@ -122,8 +124,8 @@ impl<I: ConnId> Endpoint<I> {
     ) -> Result<
         (
             Self,
-            IncomingConnections,
-            IncomingMessages,
+            IncomingConnections<I>,
+            IncomingMessages<I>,
             DisconnectionEvents,
             Option<Connection<I>>,
         ),
@@ -194,7 +196,7 @@ impl<I: ConnId> Endpoint<I> {
     pub fn new_client(
         local_addr: impl Into<SocketAddr>,
         config: Config,
-    ) -> Result<(Self, IncomingMessages, DisconnectionEvents), ClientEndpointError> {
+    ) -> Result<(Self, IncomingMessages<I>, DisconnectionEvents), ClientEndpointError> {
         let config = InternalConfig::try_from_config(config)?;
 
         let (endpoint, _, channels) =
@@ -212,7 +214,7 @@ impl<I: ConnId> Endpoint<I> {
         local_addr: SocketAddr,
         config: InternalConfig,
         builder: quinn::EndpointBuilder,
-    ) -> Result<(Self, quinn::Incoming, Channels), quinn::EndpointError> {
+    ) -> Result<(Self, quinn::Incoming, Channels<I>), quinn::EndpointError> {
         let (quic_endpoint, quic_incoming) = builder.bind(&local_addr)?;
         let local_addr = quic_endpoint
             .local_addr()
@@ -401,76 +403,6 @@ impl<I: ConnId> Endpoint<I> {
         connection.open_bi(priority).await
     }
 
-    /// Send a message to a peer over an existing connection.
-    ///
-    /// # Priority
-    ///
-    /// Locally buffered data from streams with higher priority will be transmitted before data from
-    /// streams with lower priority. Changing the priority of a stream with pending data may only
-    /// take effect after that data has been transmitted. Using many different priority levels per
-    /// connection may have a negative impact on performance.
-    ///
-    /// `0` is a sensible default for 'normal' priority.
-    ///
-    /// # Connection pooling
-    ///
-    /// Note that unlike most methods on `Endpoint`, this **will not** use the connection pool. This
-    /// method is intended to be used when it's necessary to send a message on an existing
-    /// connection only.
-    ///
-    /// # Errors
-    ///
-    /// If a connection with `dest` exists in the pool but the message fails to send,
-    /// `Err(Some(_))` will be returned. If there's no connection with `dest` in the pool, then
-    /// `Err(None)` will be returned.
-    pub async fn try_send_message(
-        &self,
-        msg: Bytes,
-        dest: &SocketAddr,
-        priority: i32,
-    ) -> Result<(), Option<SendError>> {
-        self.try_send_message_with(msg, dest, priority, None).await
-    }
-
-    /// Send a message to a peer over an existing connection using given retry configuration.
-    ///
-    /// The given `retries`, if any, will override the [`Config::retry_config`] used to create the
-    /// endpoint.
-    ///
-    /// # Priority
-    ///
-    /// Locally buffered data from streams with higher priority will be transmitted before data from
-    /// streams with lower priority. Changing the priority of a stream with pending data may only
-    /// take effect after that data has been transmitted. Using many different priority levels per
-    /// connection may have a negative impact on performance.
-    ///
-    /// `0` is a sensible default for 'normal' priority.
-    ///
-    /// # Errors
-    ///
-    /// If a connection with `dest` exists in the pool but the message fails to send,
-    /// `Err(Some(_))` will be returned. If there's no connection with `dest` in the pool, then
-    /// `Err(None)` will be returned.
-    pub async fn try_send_message_with(
-        &self,
-        msg: Bytes,
-        dest: &SocketAddr,
-        priority: i32,
-        retries: Option<&RetryConfig>,
-    ) -> Result<(), Option<SendError>> {
-        if let Some((conn, guard)) = self.connection_pool.get_by_addr(dest).await {
-            trace!("Connection exists in the connection pool: {}", dest);
-            let connection = self.wrap_connection(conn, guard);
-            retries
-                .unwrap_or(&self.config.retry_config)
-                .retry(|| async { Ok(connection.send_uni(msg.clone(), priority).await?) })
-                .await?;
-            Ok(())
-        } else {
-            Err(None)
-        }
-    }
-
     /// Close all the connections of this endpoint immediately and stop accepting new connections.
     pub fn close(&self) {
         let _ = self.termination_tx.send(());
@@ -506,18 +438,19 @@ impl<I: ConnId> Endpoint<I> {
                     .connection_pool
                     .insert(id, connection.remote_address(), connection.clone())
                     .await;
+                let connection = self.wrap_connection(connection, remover);
 
                 listen_for_incoming_messages(
+                    connection.clone(),
                     new_connection.uni_streams,
                     new_connection.bi_streams,
-                    remover.clone(),
                     self.message_tx.clone(),
                     self.disconnection_tx.clone(),
                     self.clone(),
                 );
 
                 let _ = completion.complete(Ok(()));
-                Ok(self.wrap_connection(connection, remover))
+                Ok(connection)
             }
             Err(error) => {
                 let _ = completion.complete(Err(error.clone()));
@@ -671,7 +604,7 @@ impl<I: ConnId> Endpoint<I> {
     }
 
     /// Wrap a quinn connection, setting the default retry config and pool remover.
-    fn wrap_connection(
+    pub(crate) fn wrap_connection(
         &self,
         connection: quinn::Connection,
         remover: ConnectionRemover<I>,
@@ -681,15 +614,15 @@ impl<I: ConnId> Endpoint<I> {
 }
 
 // a private helper struct for passing a bunch of channel-related things
-type Msg = (SocketAddr, Bytes);
-struct Channels {
-    connection: (MpscSender<SocketAddr>, MpscReceiver<SocketAddr>),
-    message: (MpscSender<Msg>, MpscReceiver<Msg>),
+type Msg<I> = (Connection<I>, Bytes);
+struct Channels<I: ConnId> {
+    connection: (MpscSender<Connection<I>>, MpscReceiver<Connection<I>>),
+    message: (MpscSender<Msg<I>>, MpscReceiver<Msg<I>>),
     disconnection: (MpscSender<SocketAddr>, MpscReceiver<SocketAddr>),
     termination: (Sender<()>, broadcast::Receiver<()>),
 }
 
-impl Channels {
+impl<I: ConnId> Channels<I> {
     fn new() -> Self {
         Self {
             connection: mpsc::channel(STANDARD_CHANNEL_SIZE),
